use std::{
    collections::HashMap,
    fs,
    path::{Path, PathBuf},
};

use anyhow::{anyhow, Context};
use log::{debug, error};
use serde::Deserialize;
use serde_with::{serde_as, KeyValueMap};

mod enumeration;

#[derive(Deserialize)]
pub struct Show {
    #[serde(rename = "$key$")]
    pub title: String,
    pub path: PathBuf,
    #[serde(default)]
    pub tags: Vec<String>,
    #[serde(default)]
    pub parts: HashMap<u32, String>,
    #[serde(default = "default_weight")]
    pub weight: f32,
}

fn default_weight() -> f32 {
    1.0
}

pub type Shows = Vec<Show>;

#[serde_as]
#[derive(Deserialize)]
struct ShowsWrapper(#[serde_as(as = "KeyValueMap<_>")] Shows);

#[derive(Clone, Copy, Debug, Eq, Hash, PartialEq)]
pub enum EpisodeNumber {
    Standalone,
    SingleSeason(u32),
    MultiSeason(u32, u32),
}

type Episodes = HashMap<EpisodeNumber, PathBuf>;

pub fn load<P: AsRef<Path>>(shows_file: P) -> anyhow::Result<Shows> {
    Ok(serde_yaml::from_reader::<_, ShowsWrapper>(
        fs::File::open(shows_file).context("Failed to open shows file")?,
    )
    .context("Failed to parse YAML from shows file")?
    .0)
}

pub fn display_show_episode(show: &Show, episode: EpisodeNumber) -> String {
    match episode {
        EpisodeNumber::Standalone => show.title.to_string(),
        EpisodeNumber::SingleSeason(n) => format!("{} episode {}", show.title, n),
        EpisodeNumber::MultiSeason(season, ep) => format!(
            "{} {} episode {}",
            show.title,
            show.parts
                .get(&season)
                .unwrap_or(&format!("season {}", season)),
            ep
        ),
    }
}

impl Show {
    pub fn episodes(&self) -> anyhow::Result<Episodes> {
        let path = &self.path;
        let metadata =
            fs::metadata(path).context("Failed to stat the show's path to determine file type")?;
        if metadata.is_file() {
            debug!("{} is a file, standalone", path.display());
            Ok(HashMap::from([(
                EpisodeNumber::Standalone,
                path.to_path_buf(),
            )]))
        } else if metadata.is_dir() {
            debug!("{} is a directory, enumerating episodes", path.display());
            let files: Vec<PathBuf> = fs::read_dir(path)
                .context("Failed to get a directory listing for the show's path")?
                .map(|entry| {
                    let entry =
                        entry.context("Failed to read a directory entry under the show's path")?;
                    if !entry
                        .file_type()
                        .with_context(|| {
                            format!(
                                "Failed to get file type for directory entry {}",
                                entry.path().display()
                            )
                        })?
                        .is_file()
                    {
                        debug!("Skipping {}, not a file", entry.path().display());
                        return Ok(None);
                    }
                    if entry.file_name().into_string().is_err() {
                        error!(
                            "Path {} contains invalid unicode, skipping",
                            entry.path().display()
                        );
                        return Ok(None);
                    }
                    Ok(Some(entry.path()))
                })
                .filter_map(|r| r.transpose())
                .collect::<anyhow::Result<Vec<PathBuf>>>()?;
            enumeration::enumerate_episodes(files)
                .ok_or(anyhow!("Could not detect any episode numbering scheme"))
        } else {
            Err(anyhow!("The show's path is not a file or a directory"))
        }
    }
}
